import gtk
import sys
import time
import os
import threading
import thread
import math
sys.path.append('/opt/tuxdroid/api/python')
from tux import *

class Graph(gtk.DrawingArea):
    def __init__(self):
        super(Graph, self).__init__()        
        self.connect("expose_event", self.expose)
        self.context = None
        self.battery = ' '
        self.plugged = ' '
        self.charging = ' '
        
    def expose(self, widget, event):
        #when an event occur, update the context
        self.context = widget.window.cairo_create()
        self.context.rectangle(event.area.x, event.area.y,
                               event.area.width, event.area.height)
        self.context.clip()
        self.draw()
        return False
        
    def refresh(self):
        if self.window:
            width,height = self.window.get_size()
            self.window.invalidate_rect(gtk.gdk.Rectangle(0,0,width,height),False)
    
    def draw(self):
        self.rect = self.get_allocation()
    
        self.draw_space()
        self.draw_hor_div()
        self.draw_ver_div()
        self.draw_threshold()
        self.draw_label()
        self.draw_curve()

    def draw_space(self):
     #draw axes
        #draw a rectangle in the window area
        self.context.rectangle(self.rect.x, self.rect.y, self.rect.width, self.rect.height)
        #set a white background
        self.context.set_source_rgb(1, 1, 1)
        self.context.fill()
        #draw a black border
        self.context.set_source_rgb(0, 0, 0)
        self.context.set_line_width(1)
        self.context.stroke()        
        
        
    def draw_hor_div(self):
    #draw horizontal grid with 10 div. and 10 sub-div
        for i in range(20):
            y = (self.rect.height / 20.0) * float(i)
            #place correctly the vector
            self.context.move_to(self.rect.width, y)
            if i % 2 == 1:
                #Minor div: no text
                self.context.set_source_rgb(0.2, 0.2, 0.2)
                self.context.set_line_width(0.4)
                self.context.line_to(0, y)
            else:
                #Major div: place text
                self.context.set_source_rgb(0, 0, 0)
                self.context.set_line_width(0.6)
                self.context.line_to(0, y)
                self.context.set_source_rgb(0.2, 0.2, 0.2)
                if i != 18:
                    self.context.show_text(str(6.5 - (0.125 * i))+' V')
            self.context.stroke()
            
    def draw_ver_div(self):
        #draw vertical division
        #create a tab with the wanted divisions 
        temp_tab = [0.25, 0.5, 0.75]
        for i in range(3):
            self.context.set_source_rgb(0.2, 0.2, 0.2)
            self.context.set_line_width(0.6)
            #place correctly the vector
            self.context.move_to(self.rect.width * temp_tab[i], 0)
            #trace a line
            self.context.line_to(self.rect.width * temp_tab[i], self.rect.height)
            #replace the vector to draw text
            self.context.move_to(self.rect.width * temp_tab[i], self.rect.height - 5)
            #determine the time labels
            txt = ((self.measure_graph.rate * self.measure_graph.sample)/60 - (self.measure_graph.rate * self.measure_graph.sample * temp_tab[i])/60)
            txt = str('%.2f' % txt)
            self.context.show_text(str(txt)  + " min.")
            self.context.stroke()
            
    def draw_threshold(self):
        coeff_y = float(self.rect.height / self.measure_graph.RANGE)
        
        y = self.measure_graph.convert_value([0, 1, 0, self.measure_graph.low_threshold])
        y = (y * 3.4)
        
        self.context.set_source_rgb(0.8, 0, 0)
        self.context.set_line_width(1)
        
        self.context.move_to(self.rect.width, self.rect.height - y * coeff_y)
        self.context.line_to(0, self.rect.height - y * coeff_y)
        self.context.stroke()
        
        y = self.measure_graph.convert_value([0, 1, 0, self.measure_graph.high_threshold])
        y = (y * 3.4)
        self.context.set_source_rgb(0, 0.8, 0)
        self.context.set_line_width(1)
        self.context.move_to(self.rect.width, self.rect.height - y * coeff_y)
        self.context.line_to(0, self.rect.height - y * coeff_y)
        
        self.context.stroke()

    def draw_label(self):
    #display sample and rate      
        #move the vector in the bottom
        self.draw_circle(5, self.rect.height - 7, [0.2, 0.2, 0.7])
        self.context.set_source_rgb(0.2, 0.2, 0.2)
        self.context.move_to(10, self.rect.height - 2)
        self.context.show_text(self.battery)
        
        self.draw_square(5, self.rect.height - 20, [0.2, 0.2, 0.7])
        self.context.set_source_rgb(0.2, 0.2, 0.2)
        self.context.move_to(10, self.rect.height - 15)
        self.context.show_text(self.plugged)

        self.draw_triangle(5, self.rect.height - 33, [0.2, 0.2, 0.7])
        self.context.set_source_rgb(0.2, 0.2, 0.2)
        self.context.move_to(10, self.rect.height - 28)
        self.context.show_text(self.charging)
        
        self.context.stroke()
        
    def draw_curve(self):
        coeff_y = float(self.rect.height / self.measure_graph.RANGE)
        coeff_x = float(self.rect.width / (self.measure_graph.sample - 1))
        
        self.fill(coeff_x, coeff_y)
        
        #and trace the curce
        self.context.set_line_width(0.6)
        self.context.set_source_rgb(0, 0, 1)
        self.context.move_to(self.rect.x - 2, self.rect.height + 2)
        self.context.line_to(self.rect.x, self.rect.height - (self.measure_graph.val_table[0] * coeff_y))
        temp_event = 0
        for i, val in enumerate(self.measure_graph.val_table):
            self.context.line_to(self.rect.x + ((i) * coeff_x), self.rect.height - (val * coeff_y))
            
            if self.measure_graph.val_table_event[i] == 0:
                self.draw_circle(self.rect.x + ((i) * coeff_x), self.rect.height - (val * coeff_y), [0.2, 0.2, 0.7])
            if self.measure_graph.val_table_event[i] == 1:
                self.draw_square(self.rect.x + ((i) * coeff_x), self.rect.height - (val * coeff_y), [0.2, 0.2, 0.7])
            if self.measure_graph.val_table_event[i] == 2:
                self.draw_triangle(self.rect.x + ((i) * coeff_x), self.rect.height - (val * coeff_y), [0.2, 0.2, 0.7])      
        self.context.line_to(self.rect.x + ((i) * coeff_x), self.rect.height + 5)
        self.context.stroke()

    def draw_square(self, x, y, color):
        self.context.stroke()
        self.context.set_source_rgb(color[0], color[1], color[2])
        self.context.move_to(x - 4, y - 4)
        self.context.rectangle(x - 4, y - 4, 8, 8)
        self.context.move_to(x, y)
        self.context.fill()
        self.context.set_source_rgb(0, 0, 1)
        
    def draw_triangle(self, x, y, color):
        self.context.stroke()
        self.context.set_source_rgb(color[0], color[1], color[2])
        self.context.move_to(x, y - 5)
        self.context.line_to(x + 5, y + 5)
        self.context.line_to(x - 5, y + 5)
        self.context.line_to(x, y - 5)
        self.context.close_path()
        self.context.fill()
        self.context.move_to(x, y)
        self.context.set_source_rgb(0, 0, 1)
        
    def draw_circle(self, x, y, color):
        self.context.stroke()
        self.context.set_source_rgb(color[0], color[1], color[2])
        self.context.arc(x, y, 4, 0, 2 * math.pi)
        self.context.close_path()
        self.context.fill()
        self.context.move_to(x, y)
        self.context.set_source_rgb(0, 0, 1)  
        
    def fill(self, coeff_x, coeff_y):
    #determine the coefficients to adjust the curve in the window space
        #Fill...
        self.context.set_line_width(0.6)
        self.context.set_source_rgb(0, 0, 1)
        self.context.move_to(self.rect.x - 2, self.rect.height + 2)
        self.context.line_to(self.rect.x, self.rect.height - (self.measure_graph.val_table[0] * coeff_y))
        for i, val in enumerate(self.measure_graph.val_table):
            self.context.line_to(self.rect.x + ((i) * coeff_x), self.rect.height - (val * coeff_y))
        self.context.line_to(self.rect.x + ((i) * coeff_x), self.rect.height + 5)
        self.context.close_path()
        self.context.set_source_rgba(0, 0, 1, 0.2)
        self.context.fill()


class Measure(object):
    def __init__(self, sample, rate):
        self.rate = rate
        self.rate_backup = self.rate
        self.sample = sample
        self.sample_backup = self.sample
        self.val_table = [-2] * int(self.sample)
        self.val_table_event = [-1] * int(self.sample)
        self.charger_event_backup = 3
        self.test_tab = [0]
        self.absolute_threshold_low_flag = False
        self.absolute_threshold_high_flag = False
        self.low_threshold_flag = False
        self.high_threshold_flag = False
        self.end_of_charge_flag = False
        self.chkMotors = False
        self.disable_event = False
        

        self.tick = time.time() + 1
        self.average = 0
        self.count = 0
        self.GUI_on = False  
        self.start_flag = True
        self.wait = time.time() + 2.0
        self.RANGE = 340.0
        self.low_threshold = 5.0
        self.high_threshold = 6.0

        self.__on_battery_mutex = threading.Lock()
        self.__timeout_mutex = threading.Lock()
        
        
    ###########################################################
    #
    #   Events control
    #
    ###########################################################
    def event_control(self):
        if self.GUI_on:        
            self.__tab = self.test_tab
        else:  
            self.__tab = self.event_tab
       
        for i in range (len(self.__tab)):
            val = self.convert_value(self.__tab[i])
            self.Threshold_control(i, val)
            self.Change_control(i, val)

        if self.GUI_on:
            self.test_tab = self.__tab
        else:  
            self.event_tab = self.__tab  

        self.low_threshold_flag = self.absolute_threshold_low(self.low_threshold, self.low_threshold + 0.1)
        self.high_threshold_flag = self.absolute_threshold_high(self.high_threshold, self.high_threshold - 0.1)
         
        # To prevent notifications at startup
        if self.start_flag == True or self.disable_event == True:
            self.start_flag = False
            self.test_tab[0][7] = False
            self.disable_event = False
            self.start_flag = False
            self.low_threshold_flag = False
            self.high_threshold_flag = False
            for i in range (len(self.event_tab)):
                self.event_tab[i][7] = False
                
    def convert_value(self, table):
        if table[1] == 1:
        #threshold event
            value = table[3]
            #100% in the graph = 2.5V
            value = (100 * (value - 4)) / 2.5
            return value
        else:
            value = table[3]
            #100% in the graph = 2.5V
            value = (100 * (value)) / 2.5
            return value
            
    ###########################################################
    #
    #   Absolute threshold control
    #
    ###########################################################
    def absolute_threshold_low(self, value, value_off):

        if self.absolute_threshold_low_flag == False:
            val = self.convert_value([0, 1, 0, value])     
            val_act = (self.val_table[len(self.val_table)-1]*100) / self.RANGE
            val_prev = (self.val_table[len(self.val_table)-2]*100) / self.RANGE
            if val_act < val:
                self.absolute_threshold_low_flag = True
                return True
        else:
            val = self.convert_value([0, 1, 0, value_off])
            val_act = (self.val_table[len(self.val_table)-1]*100) / self.RANGE
            val_prev = (self.val_table[len(self.val_table)-2]*100) / self.RANGE
            if val_act > val:
                self.absolute_threshold_low_flag = False
        return False
        
    def absolute_threshold_high(self, value, value_off):
        if self.absolute_threshold_high_flag == False:
            val = self.convert_value([0, 1, 0, value])     
            val_act = (self.val_table[len(self.val_table)-1]*100) / self.RANGE
            val_prev = (self.val_table[len(self.val_table)-2]*100) / self.RANGE
            if val_act > val:
                self.absolute_threshold_high_flag = True
                return True
        else:
            val = self.convert_value([0, 1, 0, value_off])
            val_act = (self.val_table[len(self.val_table)-1]*100) / self.RANGE
            val_prev = (self.val_table[len(self.val_table)-2]*100) / self.RANGE
            if val_act < val:
                self.absolute_threshold_high_flag = False
        return False
                

    ###########################################################
    #
    #   Threshold control
    #
    ###########################################################
    def Threshold_control(self, index, val):
        if self.__tab[index][1] == 1:
            if self.__tab[index][7] != True:  
                if self.__tab[index][2] == 0:
                #Increasing threshold control
                    if self.Increase_control(val) == True:
                        self.__tab[index][7] = True          
                elif self.__tab[index][2] == 1:
                #Decreasing threshold control
                    if self.Decrease_control(val) == True:
                        self.__tab[index][7] = True          
                elif self.__tab[index][2] == 2:
                #Both direction theshold control
                    if self.Both_control(val) == True:
                        self.__tab[index][7] = True
    
    def Increase_control(self, value):
        val_act = (self.val_table[len(self.val_table)-1]*100) / self.RANGE
        val_prev = (self.val_table[len(self.val_table)-2]*100) / self.RANGE
        if val_prev <= value and val_act > value:
            return True

            
    def Decrease_control(self, val):
        val_act = (self.val_table[len(self.val_table)-1]*100) / self.RANGE
        val_prev = (self.val_table[len(self.val_table)-2]*100) / self.RANGE
        if val_prev >= val and val_act < val:
            return True
            
    def Both_control(self, val):
        val_act = (self.val_table[int(len(self.val_table)-1)]*100) / self.RANGE
        val_prev = (self.val_table[int(len(self.val_table)-2)]*100) / self.RANGE
        if val_prev <= val and val_act > val or val_prev >= val and val_act < val:
            return True
          
            
    ###########################################################
    #
    #   Changes control
    #
    ###########################################################
    def Change_control(self, i, val):
        #Changements control
        if self.__tab[i][1] == 0:
            pts_per_sec = self.sample / (self.rate * self.sample)
            time = int(self.__tab[i][4] * pts_per_sec)
            if self.__tab[i][7] != True:
                # Drop change part :
                if self.__tab[i][2] == 0:
                    if self.Drop_control(time, val, self.__tab[i][8], i) == True:
                        self.__tab[i][7] = True
                # Raise change part
                if self.__tab[i][2] == 1:
                    if self.Raise_control(time, val, self.__tab[i][8], i) == True:
                        self.__tab[i][7] = True
                # Drop and Raise change               
                if self.__tab[i][2] == 2:
                    if self.ChangeD_control(time, val, self.__tab[i][8], i) == True:
                        self.__tab[i][7] = True
                # Raise and Drop change                        
                if self.__tab[i][2] == 3:
                    if self.ChangeU_control(time, val, self.__tab[i][8], i) == True:
                        self.__tab[i][7] = True
        

    def Drop_control(self, time, val, timeout, index):
        #timeout control : to prevent multiple notifications with the same event
        if timeout != 0:
            timeout -= 1    
        init_val = len(self.val_table) - 1 - int(time) + timeout
        maximum = max(self.val_table[init_val:len(self.val_table)-1])
        
        if maximum >= self.val_table[len(self.val_table)-1] + ((val/100)*self.RANGE):
            self.__tab[index][8] = int(time)
            return True
        else:
            self.__tab[index][8] = timeout
                
    def Raise_control(self, time, val, timeout, index):
        if timeout != 0:
            timeout -= 1              
        init_val = len(self.val_table) - 1 - int(time) + timeout 
        minimum = min(self.val_table[init_val:(len(self.val_table)-1)])
        if minimum <= self.val_table[len(self.val_table)-1] - ((val/100)*self.RANGE):
            self.__tab[index][8] = int(time)
            return True
        else:
            self.__tab[index][8] = timeout
            
    def ChangeD_control(self, time, val, timeout, index):
        if timeout != 0:
            timeout -= 1   
        init_val = len(self.val_table) - 1 - int(time) + timeout
        maximum = max(self.val_table[init_val:len(self.val_table)-1])  
        for i in range(int(time - timeout)):
            if self.val_table[int(len(self.val_table)-1-i)] <= self.val_table[len(self.val_table) - 1] - ((val/100)*self.RANGE):
                maximum = max(self.val_table[init_val:len(self.val_table) - i])
                if maximum >= self.val_table[len(self.val_table) - 1 - i] + val:
                    self.__tab[index][8] = int(time)
                    return True
        self.__tab[index][8] = timeout
        return False
        
            
    def ChangeU_control(self, time, val, timeout, index):
        if timeout != 0:
            timeout -= 1
        init_val = len(self.val_table) - 1 - int(time) + timeout
        maximum = max(self.val_table[init_val:len(self.val_table)-1])
        for i in range(int(time - timeout)):
            if self.val_table[int(len(self.val_table)-1-i)] >= self.val_table[len(self.val_table) - 1] + ((val/100)*self.RANGE):
                minimum = min(self.val_table[init_val:len(self.val_table)-i])
                if minimum <= self.val_table[len(self.val_table) - 1 - i] - val:
                    self.__tab[index][8] = int(time)
                    return True
        self.__tab[index][8] = timeout
        return False
        
                 
    ###########################################################
    #
    #   Test functions
    #
    ###########################################################
    def test_notification(self):
        if self.test_tab[0][7] == True:
            self.test_tab[0][7] = False
            if self.test_tab[0][5] == True:
                path = self.test_tab[0][6]
                path = path.replace('%20', ' ')
                if os.path.isfile(path) and os.path.exists(path):
                    path = self.test_tab[0][6]
                    path = "'%s'" % path
                    tux.sys.shell_free(path)
                elif self.test_tab[0][1] == 0:
                    thread.start_new_thread(self.test_default_behavior, (self.changetext,))
                elif self.test_tab[0][1] == 1:
                    thread.start_new_thread(self.test_default_behavior, (self.thresholdtext,))

    def test_default_behavior(self, txt):
        tux.cmd.mouth_open()
        tux.tts.speak(txt)
        tux.cmd.mouth_close()

    ###########################################################
    #
    #   Acquire
    #
    ###########################################################
    def on_battery_level(self, args):
        if self.__on_battery_mutex.locked():
            return
        try:
            self.__on_battery_mutex.acquire()

            if not (self.chkMotors == True and args[2] == 1):
                self.average += (args[0] * 256) + args[1]
                self.count += 1
            else:
                self.disable_event = True
            
            # Refresh the sample and rate value    
            if self.sample_backup != self.sample:
                self.sample = self.sample_backup
                self.val_table = [-2] * int(self.sample) 
                self.val_table_event = [-1] * int(self.sample)
                self.charger_event_backup = 3
            if self.rate_backup != self.rate:
                self.rate_has_changed()
            
            #control the time            
            if self.tick <= time.time():
                self.tick = time.time() + self.rate
                
                #If no value has been received, use the previous.
                if self.count == 0:
                    val = self.val_table[len(self.val_table)-1]
                    
                #Else, calcul the average
                else:
                    val = self.average / self.count
                    val = val - 530
                    
                charger_event = -1
                #Event control
                if tux.status.power_plug():
                    charger_event = 1
                    if tux.status.charger_state():
                        charger_event = 2
                elif self.charger_event_backup > 0:
                    charger_event = 0         
                        
                if self.charger_event_backup != charger_event:
                    # End of the charge cycle
                    if self.charger_event_backup == 2 and charger_event == 1:
                        self.end_of_charge_flag = True
                    self.charger_event_backup = charger_event
                else:
                    charger_event = -1
                
                #Add the new value in the table and reset all flags    
                self.val_table.append(val)
                self.val_table.pop(0)
                self.val_table_event.append(charger_event)
                self.val_table_event.pop(0)
                self.average = 0
                self.count = 0
                
                #If widget GUI is displayed, refresh the graph.
                if self.GUI_on:
                    self.graph_measure.refresh()
                #And check for events
                self.event_control()
            self.__on_battery_mutex.release()
        except:
            self.__on_battery_mutex.release()
            print sys.exc_info()



    def rate_has_changed(self):
        self.rate = self.rate_backup
        pts_per_sec = self.sample / (self.rate * self.sample)
        for i in range(len(self.event_tab)):
            if self.event_tab[i][1] == 0:
                if self.event_tab[i][4] <= self.rate:
                    self.event_tab[i][4] = self.rate
                elif self.event_tab[i][4] >= self.rate * self.sample:
                    self.event_tab[i][4] = self.rate * self.sample 
            self.event_tab[i][8] = int(self.event_tab[i][4] * pts_per_sec)
        if self.test_tab[0][1] == 0:
            if self.test_tab[0][4] <= self.rate:
                self.test_tab[0][4] = self.rate       
            elif self.test_tab[0][4] >= self.rate * self.sample:
                self.test_tab[0][4] = self.rate * self.sample
            
            self.test_tab[0][8] = int(self.test_tab[0][4] * pts_per_sec)     
        self.test_tab[0][8] = int(self.test_tab[0][4] * pts_per_sec)


